/* -*-c++-*- OpenSceneGraph - Copyright (C) 1998-2006 Robert Osfield
 *
 * This library is open source and may be redistributed and/or modified under
 * the terms of the OpenSceneGraph Public License (OSGPL) version 0.0 or
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * OpenSceneGraph Public License for more details.
*/

#ifndef OSG_LOD
#define OSG_LOD 1

#include <osg/Group>

namespace osg {

/** LOD - Level Of Detail group node which allows switching between children
    depending on distance from eye point.
    Typical uses are for load balancing - objects further away from
    the eye point are rendered at a lower level of detail, and at times
    of high stress on the graphics pipeline lower levels of detail can
    also be chosen by adjusting the viewers's Camera/CullSettings LODScale value.
    Each child has a corresponding valid range consisting of a minimum
    and maximum distance. Given a distance to the viewer (d), LOD displays
    a child if min <= d < max. LOD may display multiple children simultaneously
    if their corresponding ranges overlap. Children can be in any order,
    and don't need to be sorted by range or amount of detail. If the number of
    ranges (m) is less than the number of children (n), then children m+1 through
    n are ignored.
*/
class TessellationLOD : public Group
{
    public :

        TessellationLOD();

		TessellationLOD(float closestLevel, float highestDistance);

        /** Copy constructor using CopyOp to manage deep vs shallow copy.*/
        TessellationLOD(const TessellationLOD&,const CopyOp& copyop=CopyOp::SHALLOW_COPY);

        META_Node(osg, TessellationLOD);

        typedef osg::BoundingSphere::vec_type vec_type;
        typedef osg::BoundingSphere::value_type value_type;

        virtual void traverse(NodeVisitor& nv);

/*
        virtual bool addChild(Node *child);

        virtual bool addChild(Node *child, float min, float max);

        virtual bool removeChildren(unsigned int pos,unsigned int numChildrenToRemove=1);*/

        typedef std::pair<float,float>  MinMaxPair;
        typedef std::vector<MinMaxPair> RangeList;
		typedef std::map<std::string, Vec3d> CustomCenters;

        /** Modes which control how the center of object should be determined when computing which child is active.*/
        enum CenterMode
        {
            USE_BOUNDING_SPHERE_CENTER,
            USER_DEFINED_CENTER,
            UNION_OF_BOUNDING_SPHERE_AND_USER_DEFINED
        };

//         /** Set how the center of object should be determined when computing which child is active.*/
//         void setCenterMode(CenterMode mode) { _centerMode=mode; }
// 
//         /** Get how the center of object should be determined when computing which child is active.*/
//         CenterMode getCenterMode() const { return _centerMode; }
// 
//         /** Sets the object-space point which defines the center of the osg::LOD.
//             center is affected by any transforms in the hierarchy above the osg::LOD.*/
//         inline void setCenter(const vec_type& center) { if (_centerMode!=UNION_OF_BOUNDING_SPHERE_AND_USER_DEFINED) { _centerMode=USER_DEFINED_CENTER; } _userDefinedCenter = center; }
// 
//         /** return the LOD center point. */
//         inline const vec_type& getCenter() const { if ((_centerMode==USER_DEFINED_CENTER)||(_centerMode==UNION_OF_BOUNDING_SPHERE_AND_USER_DEFINED)) return _userDefinedCenter; else return getBound().center(); }


        /** Set the object-space reference radius of the volume enclosed by the LOD.
          * Used to determine the bounding sphere of the LOD in the absence of any children.*/
        inline void setRadius(value_type radius) { _radius = radius; }

        /** Get the object-space radius of the volume enclosed by the LOD.*/
        inline value_type getRadius() const { return _radius; }

//         /** Modes that control how the range values should be interpreted when computing which child is active.*/
//         enum RangeMode
//         {
//             DISTANCE_FROM_EYE_POINT,
//             PIXEL_SIZE_ON_SCREEN
//         };
// 
//         /** Set how the range values should be interpreted when computing which child is active.*/
//         void setRangeMode(RangeMode mode) { _rangeMode = mode; }
// 
//         /** Get how the range values should be interpreted when computing which child is active.*/
//         RangeMode getRangeMode() const { return _rangeMode; }


//         /** Sets the min and max visible ranges of range of specific child.
//             Values are floating point distance specified in local objects coordinates.*/
//         void setRange(unsigned int childNo, float min,float max);
// 
//         /** returns the min visible range for specified child.*/
//         inline float getMinRange(unsigned int childNo) const { return _rangeList[childNo].first; }
// 
//         /** returns the max visible range for specified child.*/
//         inline float getMaxRange(unsigned int childNo) const { return _rangeList[childNo].second; }
// 
//         /** returns the number of ranges currently set.
//           * An LOD which has been fully set up will have getNumChildren()==getNumRanges(). */
//         inline unsigned int getNumRanges() const { return _rangeList.size(); }
// 
//         /** set the list of MinMax ranges for each child.*/
//         inline void setRangeList(const RangeList& rangeList) { _rangeList=rangeList; }
// 
//         /** return the list of MinMax ranges for each child.*/
//         inline const RangeList& getRangeList() const { return _rangeList; }

        virtual BoundingSphere computeBound() const;

		inline const char * getDefaultTessLevel() { return "DefaultTessLevel";}

		inline void setLevel(float closestLevel) { _closestLevel = closestLevel; _decreaseRate = closestLevel/_highestDistance;}

        inline float getLevel() const { return _closestLevel; }

		inline void setDistance(float highestDistance) { _highestDistance = highestDistance; _decreaseRate = _closestLevel/highestDistance;}

		inline float getDistance() const { return _highestDistance; }

		inline void addCustomCenterUniform(std::string uniformName, Vec3d center) {_centerUniformMap[uniformName] = center;};

		inline float getCurrentLevel(float distance) { return distance>_highestDistance?1:_closestLevel-_decreaseRate*distance;}

    protected :
        virtual ~TessellationLOD() {}

        CenterMode                      _centerMode;
        vec_type                        _userDefinedCenter;
        value_type                      _radius;

		CustomCenters					_centerUniformMap;

		float							_closestLevel;
		float							_decreaseRate;
		float							_highestDistance;

};

}

#endif
